#include <stdlib.h>

#include "queue.h"

// Queue node structure
typedef struct QueueNode_ {
	int value;
	struct QueueNode_* next;
} QueueNode;

// Queue structure
struct Queue_ {
	QueueNode* head;
	QueueNode* tail;
	int size;
};

// Variables for global free node list
QueueNode* freeQueueNodes = NULL;
int nFreeQueueNodes = 0;
const int MAX_FREE_QUEUE_NODES = 5000;

/**
 * Create/allocate a new integer queue.
 */
Queue* createQueue() {
	Queue* q = malloc(sizeof(Queue));
	q->head = NULL;
	q->tail = NULL;
	q->size = 0;
	return q;
}

/**
 * Free/deallocate a queue.
 */
void freeQueue(Queue* q) {
	while (q->size > 0)
		dequeue(q);
	free(q);
}

/**
 * Insert a new value into the end of given queue.
 */
void enqueue(Queue* q, int value) {
	QueueNode* node;
	if (nFreeQueueNodes > 0) {
		node = freeQueueNodes;
		freeQueueNodes = node->next;
		--nFreeQueueNodes;
	} else {
		node = malloc(sizeof(QueueNode));
	}
	node->value = value;
	node->next = NULL;
	if (q->head == NULL)
		q->head = node;
	else
		q->tail->next = node;
	q->tail = node;
	++(q->size);
}

/**
 * Get the number of elements in queue.
 */
int queueSize(Queue* q) {
	return q->size;
}

/**
 * Remove and get the first value in given queue. Can't handle empty queue!
 */
int dequeue(Queue* q) {
	QueueNode* node = q->head;
	q->head = q->head->next;
	int value = node->value;
	if (nFreeQueueNodes < MAX_FREE_QUEUE_NODES) {
		node->next = freeQueueNodes;
		freeQueueNodes = node;
		++ nFreeQueueNodes;
	} else {
		free(node);
	}
	--(q->size);
	return value;
}

